/*
 * Copyright (C) Youness Alaoui
 *
 * This kernel module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This kernel module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * Contributors :
 *   Youness Alaoui - For the code...
 *   George Hotz (geohot) - For the lv1_peek function
 *   CJPC - For testing and debugging
 *   Various websites - For useless code examples
 *   The kernel source code - For compensating for the inexisting docs and useless website tutorials/examples
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>

#define PROCFS_NAME             "ps3_hv_mem"

static struct proc_dir_entry *proc_entry;

/* lv1_peek function written by geohot */
volatile long lv1_peek(unsigned long real_addr) {
  unsigned long ret;
  asm volatile("mr 3, %1\n"
               "li 11, 16\n"
               "sc 1\n"
               "mr %0, 3\n"
               : "=r" (ret)
               : "r" (real_addr)
               : "r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "r11", "r12");
  return ret;
}

#define MIN(a, b) (a < b?a:b)

int hv_read(char *buffer,char **start, off_t offset, int count,
    int *eof, void *user_data)
{
  off_t i;
  long data;

  printk(KERN_INFO "procfile_read (/proc/%s) called (offset 0x%p - 0x%p) "
      "count %d\n", PROCFS_NAME,
      (void *)offset, (void *)(offset + count), count);

  /* fill the buffer, return the buffer size */
  for(i = offset; i < offset + count; i += sizeof(long)) {
    data = lv1_peek(i);
    memcpy(buffer + (i - offset), &data,
        MIN(sizeof(long), count - (i - offset)));
  }

  /* Set the start pointer to buffer so we can continuously fill up our buffer
     beyond the page size limit (4K) */
  *start = buffer;
  return count;
}

int hv_write(struct file *file, const char *buffer, unsigned long count,
                   void *data)
{
  return 0;
}

int init_module()
{
        /* create the /proc file */
        proc_entry = create_proc_entry(PROCFS_NAME, 0644, NULL);

        if (proc_entry == NULL) {
                printk(KERN_ALERT "Error: Could not initialize /proc/%s\n",
                        PROCFS_NAME);
                return -ENOMEM;
        }

        proc_entry->read_proc  = hv_read;
        proc_entry->write_proc = hv_write;
        proc_entry->mode       = S_IFREG | S_IRUGO;
        proc_entry->uid        = 0;
        proc_entry->gid        = 0;
        proc_entry->size       = 0;

        printk(KERN_INFO "/proc/%s created\n", PROCFS_NAME);
        return 0;       /* everything is ok */
}

void cleanup_module()
{
        remove_proc_entry(PROCFS_NAME, proc_entry);
        printk(KERN_INFO "/proc/%s removed\n", PROCFS_NAME);
}